#!/usr/bin/env python3
import boto3
from time import sleep
from .sts import create_boto3_client
from .account import Account, AlternateContact


class Organization:
    """Interact with AWS Organization API."""

    def __init__(self, root_ou_id, credentials):
        """Initialise the boto3 organisation session.

        :param org_client: A 'organizations' boto3 client in the Master account
        :param root_ou_id: The ID of the root Organizational Unit
        :param credentials: Tuple of (key, secret, token) of the master account
        """
        self._master_credentials = credentials
        self.root_ou_id = root_ou_id

        self._org_client = create_boto3_client(
            self.get_master_account_id(), "organizations", self._master_credentials
        )

        self._account_client = create_boto3_client(
            self.get_master_account_id(), "account", self._master_credentials
        )

    def list_organizational_units_for_parent(self, parent_ou):
        """Returns a list of OUs for the given parent."""
        organizational_units = [
            ou
            for org_units in self._org_client.get_paginator(
                "list_organizational_units_for_parent"
            ).paginate(ParentId=parent_ou)
            for ou in org_units["OrganizationalUnits"]
        ]
        return organizational_units

    def list_accounts(self):
        """Retrieves all accounts in organization."""
        existing_accounts = [
            account
            for accounts in self._org_client.get_paginator("list_accounts").paginate()
            for account in accounts["Accounts"]
        ]
        return existing_accounts

    def get_account_id(self, account_name):
        """Retrieves the account ID.

        :param account_name: The name of the account
        :return: Id of account, None if account does not exist
        """
        for account in self.list_accounts():
            if account["Name"].strip() == account_name.strip():
                return account["Id"]

        return None

    def get_ou_id(self, ou_path, parent_ou_id=None):
        """Retrieve an organisational unit id.

        :param ou_path: the name of the organisational unit. For nested ou's use / notation, eg. eu/eudev
        :param parent_ou_id: The parent from where to start parsing the ou_path, defaults to root
        :return: (ou_id, parent_id)
        """
        # Return root OU if '/' is provided
        if ou_path.strip() == "/":
            return self.root_ou_id

        # Set initial OU to start looking for given ou_path
        if parent_ou_id is None:
            parent_ou_id = self.root_ou_id

        # Parse ou_path and find the ID
        ou_hierarchy = ou_path.strip("/").split("/")
        hierarchy_index = 0

        while hierarchy_index < len(ou_hierarchy):
            org_units = self.list_organizational_units_for_parent(parent_ou_id)

            for ou in org_units:
                if ou["Name"] == ou_hierarchy[hierarchy_index]:
                    parent_ou_id = ou["Id"]
                    hierarchy_index += 1
                    break
            else:
                raise ValueError(
                    f"Could not find ou with name {ou_hierarchy} in OU list {org_units}."
                )

        return parent_ou_id

    def move_account(self, account_id, ou_path, allow_direct_move=False):
        """Move the account to an organisation unit.

        Note that we are not allowing an account to move from non-root to another ou. This
        is to avoid issues with the AWS Deployment Framework that requires you to first move
        an account to root before moving to another OU.

        :param account_id: ID of the account to move
        :param ou_path: The full path to the ou in slash notation (e.g. /europe/test )
        :param allow_direct_move: Set this to True to allow moving accounts directly between OUs
                                  without first moving to the root OU.
        """
        if ou_path != self.root_ou_id:
            ou_id = self.get_ou_id(ou_path)
        else:
            ou_id = self.root_ou_id

        # TODO error handling and check if an account always returns just one item
        response = self._org_client.list_parents(ChildId=account_id)
        source_parent_id = response["Parents"][0]["Id"]

        if source_parent_id == ou_id:
            # Account is already resided in ou_path
            return

        if (
            allow_direct_move is False
            and source_parent_id != self.root_ou_id
            and ou_id != self.root_ou_id
        ):
            raise ValueError(
                f"Trying to move an account from non-root OU to another OU. "
                "Move account to root ({self.root_ou_id}) first."
            )

        self._org_client.move_account(
            AccountId=account_id,
            SourceParentId=source_parent_id,
            DestinationParentId=ou_id,
        )

    def create_account_tags(self, account_id, tags):
        """Adds tags to given account.

        :param account_id: ID of the AWS account
        :param tags: Dict of tags to apply to account
        """
        # TODO: Make the tag definition fully declarative
        # meaning that any tag not referenced will be removed
        # from the account. use org_client.untag_resource()
        formatted_tags = [
            {"Key": key, "Value": value} for tag in tags for key, value in tag.items()
        ]
        self._org_client.tag_resource(ResourceId=account_id, Tags=formatted_tags)

    @staticmethod
    def get_master_account_id():
        """Retrieves the master account id from API."""
        org_client = boto3.client("organizations")
        response = org_client.describe_organization()
        return response["Organization"]["MasterAccountId"]

    def create_account_alias(self, account_id, account_alias):
        """Creates or updates the account alias for given account_id."""

        iam_client = create_boto3_client(account_id, "iam", self._master_credentials)

        try:
            iam_client.create_account_alias(AccountAlias=account_alias)
        except iam_client.exceptions.EntityAlreadyExistsException:
            pass  # Alias already exists

    def remove_alternate_contact(self, account_id, contact_type):
        """Removes the contact type from the account."""

        if contact_type.upper() not in ["BILLING", "OPERATIONS", "SECURITY"]:
            raise ValueError(f"Contact type {contact_type} is not supported.")

        if account_id == self.get_master_account_id():
            try:
                self._account_client.delete_alternate_contact(
                    AlternateContactType=contact_type.upper()
                )
            except Exception:
                # Alternate contact doesn't exist, but botocore exception doesn't
                # seem to be published: botocore.errorfactory.ResourceNotFoundException
                pass
        else:
            try:
                self._account_client.delete_alternate_contact(
                    AccountId=account_id, AlternateContactType=contact_type.upper()
                )
            except Exception:
                # Alternate contact doesn't exist, but botocore exception doesn't
                # seem to be published: botocore.errorfactory.ResourceNotFoundException
                pass

    def update_alternate_contact(
        self, account_id: str, contact_type: str, contact_details: AlternateContact
    ):
        """Create or update the account alternate contacts for given account.

        :param account_id: ID of the AWS account
        :param contact_type: Type of contact to update
        :param contact_details: AlternateContact class instance
        """

        if contact_type.upper() not in ["BILLING", "OPERATIONS", "SECURITY"]:
            raise ValueError(f"Contact type {contact_type} is not supported.")

        if account_id == self.get_master_account_id():
            self._account_client.put_alternate_contact(
                AlternateContactType=contact_type,
                EmailAddress=contact_details.email,
                Name=contact_details.name,
                PhoneNumber=contact_details.phone,
                Title=contact_details.title,
            )

        else:
            self._account_client.put_alternate_contact(
                AccountId=account_id,
                AlternateContactType=contact_type,
                EmailAddress=contact_details.email,
                Name=contact_details.name,
                PhoneNumber=contact_details.phone,
                Title=contact_details.title,
            )

    def create_account(self, account: Account):
        """Creates a new the account if it doesn't exist.

        :param account: Class instance of Account

        :return: account_id

            TODO: Move this to the documentation:
            # Dummy payload for account creation
                from datetime import datetime
                response = {
                    'Id': 'string',
                    'AccountName': self.full_name,
                    'State': 'SUCCEEDED',
                    'RequestedTimestamp': datetime(2015, 1, 1),
                    'CompletedTimestamp': datetime(2015, 1, 1),
                    'AccountId': '173289020951',
                    'FailureReason': 'ACCOUNT_LIMIT_EXCEEDED'
                }
        """
        if account.allow_billing is True:
            allow_billing = "ALLOW"
        else:
            allow_billing = "DENY"

        response = self._org_client.create_account(
            Email=account.email,
            AccountName=account.full_name,
            # RoleName=role_name,  # defaults to OrganizationAccountAccessRole
            IamUserAccessToBilling=allow_billing,
        )["CreateAccountStatus"]

        while response["State"] == "IN_PROGRESS":
            response = self._org_client.describe_create_account_status(
                CreateAccountRequestId=response["Id"]
            )["CreateAccountStatus"]

            if response.get("FailureReason"):
                raise IOError(
                    f"Failed to create account {account.full_name}: {response['FailureReason']}"
                )

            sleep(1)  # waiting for a sec before checking account status again

        account_id = response["AccountId"]
        # TODO: Instead of sleeping, query for the role.
        sleep(10)  # Wait until OrganizationalRole is created in new account

        return account_id



"""
# configparser.py

Module to parse and validate the yaml configuration files.
"""
import yaml
import os
from .account import Account


def read_config_files(folder):
    """Retrieve account objects from yaml configuration files in given folder.

    :param folder: Folder containing config files

    :return: list of tuples [(full_account_name, email, ou_path), ]
    """
    files = [os.path.join(folder, f) for f in os.listdir(folder)]

    accounts = []
    for filename in files:
        with open(filename, "r") as stream:
            config = yaml.safe_load(stream)

            validate_config(config["Accounts"])

            for account in config["Accounts"]:
                accounts.append(Account.load_from_config(account))

    return accounts


def validate_config(configuration):
    """Validate configuration."""
    supported_keys = [
        "OrganizationalUnitPath",
        "Email",
        "AccountFullName",
        "DeleteDefaultVPC",
        "AllowDirectMoveBetweenOU",
        "AllowBilling",
        "Alias",
        "Tags",
        "AlternateContacts",
    ]

    if not isinstance(configuration, list):
        raise ValueError(f"Configuration invalid: {configuration}")

    for account in configuration:

        # Mandatory parameters
        if "AccountFullName" not in account:
            raise ValueError(f"Missing AccountFullName in configuration: {account}")
        if "OrganizationalUnitPath" not in account:
            raise ValueError(
                f"Missing OrganizationalUnitPath in configuration: {account}"
            )
        if "Email" not in account:
            raise ValueError(f"Missing Email in configuration: {account}")

        # Optional parameters
        if "DeleteDefaultVPC" in account and not isinstance(
            account["DeleteDefaultVPC"], bool
        ):
            raise ValueError(
                f'{account["AccountFullName"]} DeleteDefaultVPC param should be Boolean'
            )

        if "AllowDirectMoveBetweenOU" in account and not isinstance(
            account["AllowDirectMoveBetweenOU"], bool
        ):
            raise ValueError(
                f'{account["AccountFullName"]} AllowDirectMoveBetweenOU param should be Boolean'
            )

        if "AllowBilling" in account and not isinstance(account["AllowBilling"], bool):
            raise ValueError(
                f'{account["AccountFullName"]} AllowBilling param should be Boolean'
            )

        if "Alias" in account and not isinstance(account["Alias"], str):
            raise ValueError(
                f'{account["AccountFullName"]} Alias param should be String'
            )

        if "AlternateContacts" in account:
            if not isinstance(account["AlternateContacts"], dict):
                raise ValueError(
                    f'{account["AlternateContacts"]} AlternateContacts param should be a dict'
                )
            for contact_type, contact_value in account["AlternateContacts"].items():
                if not isinstance(contact_value, dict):
                    raise ValueError(
                        f"AlternateContacts {contact_type} should be a dict but found {contact_value}"
                    )

        if "Tags" in account:
            if not isinstance(account["Tags"], list):
                raise ValueError(
                    f'{account["AccountFullName"]} Tags param should be a List'
                )
            for tag in account["Tags"]:
                if not isinstance(tag, dict):
                    raise ValueError(
                        f'{account["AccountFullName"]} Tags should be a Dict but found {tag}'
                    )

        # Invalid parameters
        for key in account:
            if key not in supported_keys:
                raise ValueError(f"Key {key} not supported in configuration: {account}")



"""Remove default VPC and related resources."""
import logging

logger = logging.getLogger(__name__)
from time import sleep


def retry_function(function, max_retry_seconds=180):
    """Helper function to retry describe VPC related actions.

    Retrying the given describe_* call. Sometimes the VPC service
    is not ready when you've just created a new account.

    :param function: The function to call. NOTE: don't execute it but
                     pass it like client.describe_vpcs   without the closing brackets()
    :param max_retry_seconds: Maximum amount of seconds to try

    :return: The response of the function call
    """
    while True:
        try:
            response = function()
            break
        except Exception as e:
            logger.warning(
                f"Error running {function.__name__}: {e}. Sleeping for 2 second before trying again."
            )
            max_retry_seconds - 2
            sleep(2)
            if max_retry_seconds <= 0:
                raise Exception(
                    f"Could not successfully execute function {function.__name__} "
                    f"within retry limit {max_retry_seconds}."
                )

    return response


def delete_default_vpc(client, account_id, dry_run=False):
    """Delete the default VPC in the given account id.

    :param client: EC2 boto3 client instance
    :param account_id: AWS account id
    """
    # Check and remove default VPC
    default_vpc_id = None

    vpc_response = retry_function(client.describe_vpcs)

    for vpc in vpc_response["Vpcs"]:
        if vpc["IsDefault"] is True:
            default_vpc_id = vpc["VpcId"]
            break

    if default_vpc_id is None:
        logging.info(f"No default VPC found in account {account_id}")
        return

    logging.info(f"Found default VPC Id {default_vpc_id}")
    subnet_response = retry_function(client.describe_subnets)

    default_subnets = [
        subnet
        for subnet in subnet_response["Subnets"]
        if subnet["VpcId"] == default_vpc_id
    ]

    logging.info(f"Deleting default {len(default_subnets )} subnets")
    _ = [
        client.delete_subnet(SubnetId=subnet["SubnetId"], DryRun=dry_run)
        for subnet in default_subnets
    ]

    igw_response = retry_function(client.describe_internet_gateways)
    try:
        default_igw = [
            igw["InternetGatewayId"]
            for igw in igw_response["InternetGateways"]
            for attachment in igw["Attachments"]
            if attachment["VpcId"] == default_vpc_id
        ][0]
    except IndexError:
        default_igw = None

    if default_igw:
        logging.info(f"Detaching Internet Gateway {default_igw}")
        _ = client.detach_internet_gateway(
            InternetGatewayId=default_igw, VpcId=default_vpc_id, DryRun=dry_run
        )

        logging.info(f"Deleting Internet Gateway {default_igw}")
        _ = client.delete_internet_gateway(InternetGatewayId=default_igw)

    sleep(10)  # It takes a bit of time for the dependencies to clear
    logging.info(f"Deleting Default VPC {default_vpc_id}")
    delete_vpc_response = client.delete_vpc(VpcId=default_vpc_id, DryRun=dry_run)

    return delete_vpc_response



"""Model for handling AWS accounts within the organization.

The Account class allows you to create or update a new account.
"""

from typing import List


class AlternateContact:
    """Represent an alternate contact for an AWS Account."""

    def __init__(
        self,
        contact_type: str,
        name: str = None,
        email: str = None,
        phone: str = None,
        title: str = None,
    ):
        """Initialize AlternateContact."""

        self.type = contact_type

        if name:
            self.name = name
        else:
            self.name = "NotApplicable"

        if email:
            self.email = email
        else:
            self.email = "not@applicable.com"

        if phone:
            self.phone = str(phone)
        else:
            self.phone = "0000000000"

        if title:
            self.title = title
        else:
            self.title = "NotApplicable"

    @classmethod
    def load_from_config(cls, contact_type, config):
        """Initialize Account class from configuration AlternateContacts entry."""

        if not config:
            return None

        return cls(
            contact_type=contact_type,
            name=config.get("Name"),
            email=config.get("Email"),
            phone=config.get("PhoneNumber"),
            title=config.get("Title"),
        )


class Account:
    """Represents a single AWS account."""

    def __init__(
        self,
        full_name: str,
        email: str,
        ou_path: str,
        alias: str = None,
        delete_default_vpc: bool = False,
        allow_direct_move_between_ou: bool = False,
        allow_billing: bool = True,
        update_alternate_contacts: bool = False,
        operations_contact: AlternateContact = None,
        security_contact: AlternateContact = None,
        billing_contact: AlternateContact = None,
        tags: List = None,
    ):
        """Initialize Account object.

        :param full_name: Full name of the account
        :param email: Valid email address of account
        :param ou_path: Organizational Unit path. For nested OUs use / notation, eg. us/development.
                        Use root identifier to target root OU (eg. r-abc1) or just use '/'

        :param alias: Optional Alias to use, if None the alias will default to full_name
        :param delete_default_vpc: Set this to True to delete the default vpc from the account
        :param allow_direct_move_between_ou: Set this to False to prevent the script from moving
                                             an account directly from one OU to another. This is
                                             useful if you are using the AWS Deployment Framework
                                             as it requires you to first move an account to the root
                                             to trigger the base cloudformation stack updates.
        :param update_alternate_contacts: Set this to True to update the alternate contacts
        :param allow_billing: Set this to False to prevent account admins from using the billing console
        :param operations_contact: Contact email for AWS Operations
        :param security_contact: Contact email for AWS Security
        :param billing_contact: Contact email for AWS Billing
        :param tags: a dict containing optional tags for the account
        """
        self.full_name = full_name
        self.email = email
        self.ou_path = ou_path
        self.delete_default_vpc = delete_default_vpc
        self.allow_direct_move_between_ou = allow_direct_move_between_ou
        self.allow_billing = allow_billing

        if alias is None:
            self.alias = full_name
        else:
            self.alias = alias

        self.update_alternate_contacts = update_alternate_contacts
        self.operations_contact = operations_contact
        self.security_contact = security_contact
        self.billing_contact = billing_contact

        if tags is None:
            self.tags = []
        else:
            self.tags = tags

    @classmethod
    def load_from_config(cls, config):
        """Initialize Account class from configuration object."""

        if config.get("AlternateContacts"):
            update_alternate_contacts = True
            operations_contact = AlternateContact.load_from_config(
                "OPERATIONS", config["AlternateContacts"].get("Operations")
            )
            security_contact = AlternateContact.load_from_config(
                "SECURITY", config["AlternateContacts"].get("Security")
            )
            billing_contact = AlternateContact.load_from_config(
                "BILLING", config["AlternateContacts"].get("Billing")
            )
        else:
            update_alternate_contacts = False
            operations_contact = None
            security_contact = None
            billing_contact = None

        return cls(
            config["AccountFullName"],
            config["Email"],
            config["OrganizationalUnitPath"],
            alias=config.get("Alias", None),
            delete_default_vpc=config.get("DeleteDefaultVPC", False),
            allow_direct_move_between_ou=config.get("AllowDirectMoveBetweenOU", False),
            allow_billing=config.get("AllowBilling", True),
            update_alternate_contacts=update_alternate_contacts,
            operations_contact=operations_contact,
            security_contact=security_contact,
            billing_contact=billing_contact,
            tags=config.get("Tags", []),
        )
